"""Requests module."""
from typing import Any, Tuple, Union

from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # pylint: disable=import-error

from wmflib import __version__


TypeTimeout = Union[float, Tuple[float, float]]
"""Type alias for the requests timeout parameter."""
DEFAULT_TIMEOUT: TypeTimeout = (3.0, 5.0)
""":py:const:`wmflib.requests.TypeTimeout`: the default timeout to use if none is passed, in seconds."""


class TimeoutHTTPAdapter(HTTPAdapter):
    """Requests HTTP Adapter with default timeout for all requests.

    See Also:
        https://hodovi.ch/blog/advanced-usage-python-requests-timeouts-retries-hooks/

    """

    def __init__(self, **kwargs: Any):
        """Initialize the adapter with a default timeout, that can be overriden.

        To override the default timeout of :py:const:`wmflib.requests.DEFAULT_REQUESTS_TIMEOUT`` pass the ``timeout``
        parameter when initializing an instance of this class.

        Params:
            As required by requests's HTTPAdapter:
            https://2.python-requests.org/en/master/api/#requests.adapters.HTTPAdapter

        """
        self.timeout = DEFAULT_TIMEOUT
        if 'timeout' in kwargs:
            self.timeout = kwargs['timeout']
            del kwargs['timeout']

        super().__init__(**kwargs)

    def send(self, request: PreparedRequest,  # type: ignore # pylint: disable=arguments-differ
             **kwargs: Any) -> Response:
        """Override the send method to pass the default timeout if not set.

        Params:
            As required by requests's HTTPAdapter:
            https://2.python-requests.org/en/master/api/#requests.adapters.HTTPAdapter.send
            The ``noqa`` is needed unless the exact signature is replicated.

        """
        kwargs['timeout'] = kwargs.get('timeout', self.timeout)
        return super().send(request, **kwargs)


def http_session(name: str, *, timeout: TypeTimeout = DEFAULT_TIMEOUT, tries: int = 3,
                 backoff: float = 1.0) -> Session:
    """Return a new requests Session with User-Agent, default timeout and retry logic on failure already setup.

    The returned session will retry any ``DELETE, GET, HEAD, OPTIONS, PUT, TRACE`` request that returns one of
    the following HTTP status code:

        - 429 Too Many Requests
        - 500 Internal Server Error
        - 502 Bad Gateway
        - 503 Service Unavailable
        - 504 Gateway Timeout

    It will also retry any request that times out before the specified timeout. For non-idempotent HTTP methods the
    request will not be retried if the data has reached the server.

    The retry interval between requests is determined by the ``backoff`` parameter, see below.

    The timeout functionality is provided via the :py:class:`wmflib.requests.TimeoutHTTPAdapter` and individual request
    can override the session timeout by specifying a ``timeout`` parameter.

    Examples:
        With default parameters::

            from wmflib.requests import http_session
            session = http_session('AppName')  # The given name will be used in the User-Agent header, see below
            # At this point the session can be used as a normal requests session

        With customized parameters::

            session = http_session('AppName', timeout=10.0, tries=5, backoff=2.0)

    See Also:
        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
        https://docs.python-requests.org/en/latest/user/advanced/#timeouts

    Arguments:
        name (str): the name to use for the User-Agent header. It can be specified in the ``name/version`` format, if
            applicable. The resulting header will be set to::

                pywmflib/{version} {name} +https://wikitech.wikimedia.org/wiki/Python/Wmflib root@wikimedia.org
        timeout (float, tuple): the default timeout to use in all requests within this session, in seconds. Any request
            can override it passing the ``timeout`` parameter explicitely. It can be either a single float or a tuple
            of two floats (connect, read), according to requests's documentation.
        tries (int): the total number of requests to perform before bailing out.
        backoff (float): the backoff factor to use, will generate a sleep between retries, in seconds, of::

            {backoff factor} * (2 ** ({number of total retries} - 1))

    Returns:
        requests.Session: the pre-configured session.

    """
    retry_strategy = Retry(total=tries, backoff_factor=backoff, status_forcelist=[429, 500, 502, 503, 504])
    adapter = TimeoutHTTPAdapter(timeout=timeout, max_retries=retry_strategy)
    session = Session()
    user_agent = f'pywmflib/{__version__} {name} +https://wikitech.wikimedia.org/wiki/Python/Wmflib'
    session.headers.update({'User-Agent': user_agent})
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    return session



"""Dnsdisc module tests."""
from datetime import timedelta
from unittest import mock

import pytest

from wmflib.decorators import get_backoff_sleep, retry, RetryParams
from wmflib.exceptions import WmflibError


class TestRetryParams:
    """Test class for the RetryParams dataclass."""

    def setup_method(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.base_params = {
            'tries': 3,
            'delay': timedelta(seconds=10),
            'backoff_mode': 'constant',
            'exceptions': (WmflibError,),
            'failure_message': 'failed',
        }

    def test_validate_ok(self):
        """It should validate the parameters if they are correct."""
        params = RetryParams(**self.base_params)
        params.validate()  # Will raise if invalid

    @pytest.mark.parametrize('modified, expected', (
        ({'backoff_mode': 'invalid'}, 'Invalid backoff_mode: invalid'),
        ({'backoff_mode': 'exponential', 'delay': timedelta(seconds=0.5)},
         'Delay must be greater than 1 if backoff_mode is exponential'),
        ({'tries': 0}, 'Tries must be a positive integer, got 0'),
        ({'failure_message': ''}, 'A failure_message must be set.'),
    ))
    def test_validate_invalid_params(self, modified, expected):
        """It should raise WmflibError if the parameters are invalid."""
        params = RetryParams(**{**self.base_params, **modified})
        with pytest.raises(WmflibError, match=expected):
            params.validate()


def _generate_mocked_function(calls):
    func = mock.Mock()
    func.side_effect = calls
    func.__qualname__ = 'mocked.func'
    return func


@pytest.mark.parametrize('calls, sleep_calls', (
    ([True], []),
    ([WmflibError('error1'), True], [3.0]),
    ([WmflibError('error1'), WmflibError('error2'), True], [3.0, 9.0]),
))
@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_pass_no_args(mocked_sleep, calls, sleep_calls):
    """Using @retry with no arguments should use the default values."""
    func = _generate_mocked_function(calls)
    ret = retry(func)()
    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('exc, calls, sleep_calls', (
    (WmflibError, [WmflibError('error')] * 3, [3.0, 9.0]),
    (Exception, [Exception('error')], []),
))
@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_fail_no_args(mocked_sleep, exc, calls, sleep_calls):
    """Using @retry with no arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function(calls)
    with pytest.raises(exc, match='error'):
        retry(func)()

    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('calls, sleep_calls, kwargs', (
    ([True], [], {'delay': timedelta(seconds=11), 'tries': 1}),
    ([Exception('error1'), True], [5.55],
     {'delay': timedelta(seconds=5, milliseconds=550), 'tries': 2, 'exceptions': Exception}),
    ([WmflibError('error1'), True], [8.88],
     {'backoff_mode': 'exponential', 'delay': timedelta(milliseconds=8880), 'tries': 2}),
    ([WmflibError('error1'), True], [0.90],
     {'backoff_mode': 'power', 'delay': timedelta(milliseconds=900), 'tries': 2}),
))
@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_pass_args(mocked_sleep, calls, sleep_calls, kwargs):
    """Using @retry with arguments should use the specified values."""
    func = _generate_mocked_function(calls)
    ret = retry(**kwargs)(func)()

    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('exc, kwargs', (
    (WmflibError, {}),
    (RuntimeError, {'exceptions': (KeyError, ValueError)}),
))
@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_fail_args(mocked_sleep, exc, kwargs):
    """Using @retry with arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function([exc('error')])
    kwargs['tries'] = 1
    with pytest.raises(exc, match='error'):
        retry(**kwargs)(func)()

    func.assert_called_once_with()
    assert not mocked_sleep.called


@pytest.mark.parametrize('failure_message, expected_log', (
    (None, "Attempt to run 'unittest.mock.mocked.func' raised:"),
    ('custom failure message', 'custom failure message'),
))
@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_failure_message(mocked_sleep, failure_message, expected_log, caplog):
    """Using @retry with a failure_message should log that message when an exception is caught."""
    func = _generate_mocked_function([WmflibError('error1'), True])
    ret = retry(failure_message=failure_message)(func)()  # pylint: disable=no-value-for-parameter

    assert ret
    assert expected_log in caplog.text
    assert 'error1' in caplog.text
    assert mocked_sleep.call_count == 1


@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_fail_chained_exceptions(mocked_sleep, caplog):
    """When @retry catches a chained exception, it should log exception messages all the way down the chain."""
    def side_effect():
        try:
            raise WmflibError('error2') from WmflibError('error3')
        except WmflibError:
            raise WmflibError('error1')  # pylint: disable=raise-missing-from
    func = _generate_mocked_function(side_effect)
    with pytest.raises(WmflibError, match='error1'):
        retry(func)()

    assert 'error1\nRaised while handling: error2\nCaused by: error3' in caplog.text
    assert mocked_sleep.call_count == 2


@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_retry_dynamic_params_callback(mocked_sleep):
    """It should execute the given callback and use the new parameters."""
    def callback(params, _func, _args, _kwargs):
        """Alter the tries value."""
        params.tries = 2

    func = _generate_mocked_function([WmflibError('error1'), True])
    ret = retry(tries=1, dynamic_params_callbacks=(callback,))(func)()  # pylint: disable=no-value-for-parameter

    assert ret
    assert mocked_sleep.call_count == 1


@pytest.mark.parametrize('mode, base, values', (
    ('constant', 0, (0,) * 5),
    ('constant', 0.5, (0.5,) * 5),
    ('constant', 3, (3,) * 5),
    ('linear', 0, (0,) * 5),
    ('linear', 0.5, (0.5, 1.0, 1.5, 2.0, 2.5)),
    ('linear', 3, (3, 6, 9, 12, 15)),
    ('power', 0, (0,) * 5),
    ('power', 0.5, (0.5, 1, 2, 4, 8)),
    ('power', 3, (3, 6, 12, 24, 48)),
    ('exponential', 1, (1,) * 5),
    ('exponential', 1.5, (1.5, 2.25, 3.375, 5.0625, 7.59375)),
    ('exponential', 3, (3, 9, 27, 81, 243)),
))
def test_get_backoff_sleep(mode, base, values):
    """Calling get_backoff_sleep() should return the proper backoff based on the arguments."""
    for i, val in enumerate(values, start=1):
        assert get_backoff_sleep(mode, base, i) == pytest.approx(val)


def test_get_backoff_sleep_raise():
    """Calling get_backoff_sleep() with an invalid backoff_mode should raise ValueError."""
    with pytest.raises(ValueError, match='Invalid backoff_mode: invalid'):
        get_backoff_sleep('invalid', 1, 5)



"""Prometheus module."""

import logging

from typing import Dict, List, Optional, Union

import requests

from wmflib.constants import ALL_DATACENTERS
from wmflib.exceptions import WmflibError
from wmflib.requests import http_session

logger = logging.getLogger(__name__)


class PrometheusError(WmflibError):
    """Custom exception class for errors of this module."""


class Prometheus:
    """Class to interact with the Prometheus server."""

    _prometheus_api: str = 'http://prometheus.svc.{site}.wmnet/ops/api/v1/query'

    def __init__(self) -> None:
        """Initialize the instance.

        Examples:
            ::

                >>> from wmflib.prometheus import Prometheus
                >>> prometheus = Prometheus()

        """
        self._http_session = http_session('.'.join((self.__module__, self.__class__.__name__)))

    def query(self, query: str, site: str, *, timeout: Optional[Union[float, int]] = 10) -> List[Dict]:
        """Perform a generic query.

        Examples:
            ::

                results = prometheus.query('node_memory_MemTotal_bytes{instance=~"host1001:.*"}', 'eqiad')

            The results will be something like::

                [
                    {
                        'metric': {
                            '__name__': 'node_memory_MemTotal_bytes',
                            'cluster': 'management',
                            'instance': 'host1001:9100',
                            'job': 'node',
                            'site': 'eqiad'
                        },
                        'value': [1636569623.988, '67225329664']
                    }
                ]

        Arguments:
            query (str): a prometheus query
            site (str): The site to use for queries. Must be one of
                :py:const:`wmflib.constants.ALL_DATACENTERS`
            timeout (float, int, None, optional): How many seconds to wait for the prometheus to
                send data before giving up, as a float or int. Alternatively None to indicate an
                infinite timeout.

        Returns:
            list: returns an empty list if there are no results otherwise return a list of
            results of the form: {'metric': {}, 'value': [$timestamp, $value]}

        Raises:
            wmflib.prometheus.PrometheusError: on error

        """
        if site not in ALL_DATACENTERS:
            msg = f'site ({site}) must be one of wmflib.constants.ALL_DATACENTERS {ALL_DATACENTERS}'
            raise PrometheusError(msg)

        url = self._prometheus_api.format(site=site)
        response = self._http_session.get(url, params={'query': query}, timeout=timeout)
        if response.status_code != requests.codes['ok']:
            msg = f'Unable to get metric: HTTP {response.status_code}: {response.text}'
            raise PrometheusError(msg)

        result = response.json()

        if result.get('status', 'error') == 'error':
            msg = f'Unable to get metric: {result.get("error", "unknown")}'
            raise PrometheusError(msg)

        return result['data']['result']



"""IRC utils module tests."""
import logging

from unittest import mock

from wmflib.irc import SALSocketHandler


GENERIC_LOG_RECORD = logging.LogRecord('module', logging.DEBUG, '/source/file.py', 1, 'message', [], None)


@mock.patch('socket.gethostname', return_value='current-hostname')
def test_irc_socket_handler_init(mocked_hostname):
    """An instance of SALSocketHandler should set the address for the socket and the user."""
    handler = SALSocketHandler('host', 123, 'user')
    assert handler.addr == ('host', 123)
    assert handler.username == 'user'
    assert handler.hostname == 'current-hostname'
    mocked_hostname.assert_called_once_with()


@mock.patch('wmflib.irc.socket')
def test_irc_socket_handler_emit_ok(mocked_socket):
    """Calling emit() on an SALSocketHandler instance should send the message to the socket."""
    handler = SALSocketHandler('host', 123, 'user')

    handler.emit(GENERIC_LOG_RECORD)

    mocked_socket.assert_has_calls([
        mock.call.socket().connect(('host', 123)), mock.call.socket().close()], any_order=True)


@mock.patch('wmflib.irc.socket')
def test_irc_socket_handler_emit_ko(mocked_socket):
    """If an error occur while calling emit() on an SALSocketHandler instance, it should call ."""
    handler = SALSocketHandler('host', 123, 'user')
    handler.handleError = mock.MagicMock()
    mocked_socket.socket.side_effect = OSError

    handler.emit(GENERIC_LOG_RECORD)

    assert mock.call.socket().connect(('host', 123)) not in mocked_socket.mock_calls
    assert mock.call.socket().close() not in mocked_socket.mock_calls
    handler.handleError.assert_called_once_with(GENERIC_LOG_RECORD)



"""Constants module tests."""

from wmflib.constants import ALL_DATACENTERS, CORE_DATACENTERS, DATACENTER_NUMBERING_PREFIX


def test_datacenters():
    """Verify both datacenter constants are in sync."""
    assert set(ALL_DATACENTERS) == set(DATACENTER_NUMBERING_PREFIX)


def test_core_datacenters():
    """Verify that the core datacenters is a subset of all datacenters."""
    assert set(CORE_DATACENTERS).issubset(set(ALL_DATACENTERS))



"""Interactive module tests."""
import logging

from unittest import mock

import pytest

from wmflib import interactive
from wmflib.exceptions import WmflibError
from wmflib.tests import check_logs, require_caplog


def example_division(positional: int, *, keyword: int = 1) -> int:
    """Example function to be used in the confirm_on_failure() tests.

    - When passing keyword=0 will raise a ZeroDivisionError
    - When passing 0 to the positional argument will raise a wmflib.interactive.AbortError
    - In all other cases will return positional divided by keyword

    """
    if positional == 0:
        raise interactive.AbortError('Test aborted')

    return positional / keyword


@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ask_input_ok(mocked_isatty, mocked_input, capsys):
    """Calling ask_input() should return the user input if valid."""
    valid_answer = 'valid'
    mocked_isatty.return_value = True
    mocked_input.return_value = valid_answer
    message = 'Test message'
    choice = interactive.ask_input(message, [valid_answer, 'other'])
    assert choice == valid_answer
    out, _ = capsys.readouterr()
    assert message in out
    assert 'Invalid response' not in out


@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ask_input_ko(mocked_isatty, mocked_input, capsys):
    """Calling ask_input() should raise InputError if the correct answer is not provided."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'invalid'
    message = 'Test message'
    with pytest.raises(interactive.InputError, match='Too many invalid answers'):
        interactive.ask_input(message, ['go'])

    out, _ = capsys.readouterr()
    assert message in out
    assert out.count('Invalid response') == 3


@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ask_input_no_tty(mocked_isatty):
    """It should raise InputError if not in a TTY."""
    mocked_isatty.return_value = False
    with pytest.raises(interactive.InputError, match='Not in a TTY, unable to ask for input'):
        interactive.ask_input('message', [])


@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ask_confirmation_go(mocked_isatty, mocked_input, capsys):
    """Calling ask_confirmation() should not raise if the correct answer is provided."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'go'
    message = 'Test message'
    interactive.ask_confirmation(message)
    out, _ = capsys.readouterr()
    assert message in out


@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ask_confirmation_abort(mocked_isatty, mocked_input):
    """Calling ask_confirmation() should raise AbortError if 'abort' is provided."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'abort'
    message = 'Test message'
    with pytest.raises(interactive.AbortError, match='Confirmation manually aborted'):
        interactive.ask_confirmation(message)


def test_confirm_on_failure_ok():
    """Calling confirm_on_failure() should run the given function and return its value if no exception is raised."""
    ret = interactive.confirm_on_failure(example_division, 4, keyword=2)
    assert ret == 2


@require_caplog
@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_confirm_on_failure_abort(mocked_isatty, mocked_input, capsys, caplog):
    """It should ask for input if the called function raises an exception and allow to abort it."""
    caplog.set_level(logging.INFO)
    mocked_isatty.return_value = True
    mocked_input.return_value = 'abort'
    with pytest.raises(interactive.AbortError, match='Task manually aborted'):
        interactive.confirm_on_failure(example_division, 1, keyword=0)

    out, _ = capsys.readouterr()
    assert 'What do you want to do' in out
    check_logs(caplog, 'Failed to run wmflib.tests.unit.test_interactive.example_division', logging.ERROR)


@require_caplog
def test_confirm_on_failure_func_abort(capsys, caplog):
    """It should let an AbortError exception raised in the called function pass through, not asking the user twice."""
    caplog.set_level(logging.INFO)
    with pytest.raises(interactive.AbortError, match='Test aborted'):
        interactive.confirm_on_failure(example_division, 0, keyword=1)

    out, _ = capsys.readouterr()
    assert 'What do you want to do' not in out
    log_message = 'Failed to run wmflib.tests.unit.test_interactive.example_division'
    try:
        check_logs(caplog, log_message, logging.ERROR)
        raise AssertionError(f'Found unexpected message in logs: {log_message}')
    except RuntimeError:
        pass  # No log message found, as expected


@require_caplog
@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_confirm_on_failure_skip(mocked_isatty, mocked_input, capsys, caplog):
    """It should ask for input if the called function raises an exception and allow to skip it returning None."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'skip'

    ret = interactive.confirm_on_failure(example_division, 1, keyword=0)

    assert ret is None
    out, _ = capsys.readouterr()
    assert 'What do you want to do' in out
    check_logs(caplog, 'Failed to run wmflib.tests.unit.test_interactive.example_division', logging.ERROR)


@require_caplog
@mock.patch('builtins.input')
@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_confirm_on_failure_retry(mocked_isatty, mocked_input, capsys, caplog):
    """It should ask for input if the called function raises an exception and allow to skip it returning None."""
    mocked_isatty.return_value = True
    mocked_input.side_effect = ('retry', 'retry', 'skip')

    ret = interactive.confirm_on_failure(example_division, 1, keyword=0)

    assert ret is None
    out, _ = capsys.readouterr()
    assert 'What do you want to do' in out
    check_logs(caplog, 'Failed to run wmflib.tests.unit.test_interactive.example_division', logging.ERROR)


def test_get_username_no_env(monkeypatch):
    """If no env variable is set should return '-'."""
    monkeypatch.delenv('USER', raising=False)
    monkeypatch.delenv('SUDO_USER', raising=False)
    assert interactive.get_username() == '-'


def test_get_username_root(monkeypatch):
    """When unable to detect the real user should return 'root'."""
    monkeypatch.setenv('USER', 'root')
    monkeypatch.delenv('SUDO_USER', raising=False)
    assert interactive.get_username() == 'root'


def test_get_username_ok(monkeypatch):
    """As a normal user with sudo should return the user's name."""
    monkeypatch.setenv('USER', 'root')
    monkeypatch.setenv('SUDO_USER', 'user')
    assert interactive.get_username() == 'user'


@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ensure_shell_is_durable_interactive(mocked_isatty):
    """Should raise WmflibError if in an interactive shell."""
    mocked_isatty.return_value = True
    with pytest.raises(WmflibError, match='Must be run in non-interactive mode or inside a screen or tmux.'):
        interactive.ensure_shell_is_durable()

    assert mocked_isatty.called


@mock.patch('wmflib.interactive.sys.stdout.isatty')
def test_ensure_shell_is_durable_non_interactive(mocked_isatty):
    """Should raise WmflibError if in an interactive shell."""
    mocked_isatty.return_value = False
    interactive.ensure_shell_is_durable()
    assert mocked_isatty.called


@mock.patch('wmflib.interactive.sys.stdout.isatty')
@pytest.mark.parametrize('env_name, env_value', (
    ('STY', '12345.pts-1.host'),
    ('TMUX', '/tmux-1001/default,12345,0'),
    ('TERM', 'screen-example'),
    ('TERM', 'tmux-example'),
))
def test_ensure_shell_is_durable_sty(mocked_isatty, env_name, env_value, monkeypatch):
    """Should not raise if in an interactive shell with STY set, TMUX set or a screen-line TERM."""
    mocked_isatty.return_value = True
    monkeypatch.setenv(env_name, env_value)
    interactive.ensure_shell_is_durable()
    assert mocked_isatty.called


@mock.patch('wmflib.interactive.getpass')
def test_get_secret_correct_noconfirm(mocked_getpass):
    """Should ask for secret once and return the secret."""
    mocked_getpass.getpass.return_value = 'interactive_password'
    assert interactive.get_secret('secret') == 'interactive_password'
    mocked_getpass.getpass.assert_called_once_with(prompt='secret: ')


@mock.patch('wmflib.interactive.getpass')
def test_get_secret_correct(mocked_getpass):
    """Should ask for secret twice and return the secret."""
    mocked_getpass.getpass.side_effect = ['interactive_password', 'interactive_password']
    assert interactive.get_secret('secret', confirm=True) == 'interactive_password'
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '), mock.call(prompt='Again, just to be sure: ')])


@mock.patch('wmflib.interactive.getpass')
def test_get_secret_bad_retry(mocked_getpass):
    """Should ask for secret twice and raise WmflibError if they don't match."""
    mocked_getpass.getpass.side_effect = ['interactive_password', 'foobar']
    with pytest.raises(WmflibError, match='secret: Passwords did not match'):
        interactive.get_secret('secret', confirm=True)
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '), mock.call(prompt='Again, just to be sure: ')])


@mock.patch('wmflib.interactive.getpass')
def test_get_secret_too_small(mocked_getpass):
    """Should ask for secret until the minimum length is met."""
    mocked_getpass.getpass.side_effect = ['5char', 'interactive_password']
    assert interactive.get_secret('secret') == 'interactive_password'
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '),
         mock.call(prompt='Secret must be at least 6 characters. try again: ')])



"""Prometheus module tests."""
import pytest

from wmflib.prometheus import Prometheus, PrometheusError


TEST_URI = 'http://prometheus.svc.eqiad.wmnet/ops/api/v1/query'


def get_response_data(check='ok'):
    """Return mocked data based on the test."""
    json_data = {
        'status': 'success',
        'data': {
            'resultType': 'vector',
            'result': [{
                'metric': {},
                'value': [1553173538.378, '42.42']
            }]
        }
    }
    if check == 'empty_result':
        json_data['data']['result'] = []
    elif check == 'error':
        del json_data['data']
        json_data['status'] = 'error'
        json_data['error'] = 'Foobar error'

    return json_data


class TestPrometheus:
    """Test class for the Prometheus class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.prometheus = Prometheus()

    def test_init(self):
        """It should initialise the instance."""
        assert isinstance(self.prometheus, Prometheus)

    def test_bad_site(self):
        """Test with a bad site parameter."""
        with pytest.raises(
                PrometheusError, match=r'site \(bad_site\) must be one of wmflib.constants.ALL_DATACENTERS'):
            self.prometheus.query('query_string', 'bad_site')

    def test_query_ok(self, requests_mock):
        """Check parsing a good request."""
        requests_mock.get(
            TEST_URI,
            json=get_response_data('ok'),
            status_code=200
        )
        assert 'value' in self.prometheus.query('query_string', 'eqiad')[0]

    def test_query_not_ok(self, requests_mock):
        """Check we error on a fetch failure."""
        requests_mock.get(
            TEST_URI,
            json=get_response_data('not_ok'),
            status_code=503
        )
        with pytest.raises(PrometheusError, match='Unable to get metric: HTTP 503'):
            self.prometheus.query('query_string', 'eqiad')

    def test_query_error(self, requests_mock):
        """Check handling an error in response data."""
        requests_mock.get(
            TEST_URI,
            json=get_response_data('error'),
            status_code=200
        )
        with pytest.raises(PrometheusError, match='Unable to get metric: Foobar error'):
            self.prometheus.query('query_string', 'eqiad')

    def test_query_empty_result(self, requests_mock):
        """Test for en empty result."""
        requests_mock.get(
            TEST_URI,
            json=get_response_data('empty_result'),
            status_code=200
        )
        assert not self.prometheus.query('query_string', 'eqiad')



"""Sphinx configuration checker.

- Check that all the existing wmflib modules are listed in the API index documentation.
- Check that all the modules listed in the API index documentation exists in wmflib.
- Check that all the listed modules have its own file in the api/ documentation directory.
  Sphinx would raise a warning but not fail in this case.
"""
import argparse
import pkgutil
import sys

from pathlib import Path

import wmflib


DOC_API_BASE_PATH = Path('doc/source/api')
DOC_API_INDEX_PATH = DOC_API_BASE_PATH / 'index.rst'
API_INDEX_PREFIX = '    wmflib.'
EXCLUDED_NAMES = ()


def main(base_path):
    """Perform the check."""
    wmflib_modules = {name for _, name, ispkg in pkgutil.iter_modules(wmflib.__path__)
                      if not ispkg and name not in EXCLUDED_NAMES}

    doc_path = base_path / DOC_API_INDEX_PATH
    with open(doc_path, encoding='utf-8') as f:
        api_index_lines = f.readlines()

    doc_api_lines = [line.strip() for line in api_index_lines if line.startswith(API_INDEX_PREFIX)]
    doc_api_modules = {line.split('.', 1)[1] for line in doc_api_lines}

    ret = 0
    if wmflib_modules - doc_api_modules:
        print(f'wmflib modules that are not listed in {DOC_API_INDEX_PATH}: {wmflib_modules - doc_api_modules}')
        ret += 1
    if doc_api_modules - wmflib_modules:
        print(f'Documented modules in {DOC_API_INDEX_PATH} that are missing in wmflib: '
              f'{doc_api_modules - wmflib_modules}')
        ret += 1

    doc_api_files = [f'wmflib.{name}.rst' for name in doc_api_modules]
    missing_doc_api_files = [file for file in doc_api_files
                             if not (DOC_API_BASE_PATH / file).is_file]
    if missing_doc_api_files:
        print(f'Missing documentation files in {DOC_API_BASE_PATH}: {missing_doc_api_files}')
        ret += 1

    if ret == 0:
        print('All wmflib modules are documented')

    return ret


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Check that all wmflib modules are documented')
    parser.add_argument('base_path', help='Path to the root of the wmflib repository')
    args = parser.parse_args()

    sys.exit(main(args.base_path))



"""IDM utils module tests."""
from argparse import Namespace
from unittest.mock import patch

import pytest

from wmflib.idm import IdmValueError, logoutd_args, LogoutdBase


def test_logoutd_args_no_description():
    """Test the argparsing function with no description."""
    with pytest.raises(IdmValueError, match='Must provide a string description'):
        logoutd_args()


def test_logoutd_args_list():
    """Test the argparsing function for the query action."""
    args = logoutd_args('foo', ['list'])
    assert args == Namespace(command='list', verbose=0)


def test_logoutd_args_logout():
    """Test the argparsing function for the query action."""
    args = logoutd_args('foo', ['-vv', 'query', '--cn', 'bill', '--uid', 'Bill'])
    assert args == Namespace(command='query', verbose=2, cn='bill', uid='Bill')


def test_logoutd_args():
    """Test the argparsing function for the logout action."""
    args = logoutd_args('foo', ['logout', '-u', 'bob', '--cn', 'Bob'])
    assert args == Namespace(command='logout', uid='bob', verbose=0, cn='Bob')


class ConcreteLogoutdBase(LogoutdBase):
    """Example logout class."""

    def logout_user(self, user):
        """Logout user."""
        return 'logout ' + user

    def query_user(self, user):
        """Query user."""
        return 'query ' + user

    def list(self):
        """List users."""
        return 'list'


class TestLogoutd:
    """Test the logoutd class."""

    @patch(
        'argparse.ArgumentParser.parse_args',
        return_value=Namespace(uid='uid', cn='common_name', command='query'),
    )
    def setup_method(self, _, _mock_args):
        """Configure the instance."""
        # pylint: disable=attribute-defined-outside-init
        self.idm = ConcreteLogoutdBase()

    def test_user(self):
        """By Default the user identifier should return the common_name."""
        assert self.idm.user == 'common_name'

    def test_user_identifier_uid(self):
        """Update the user identifier to use the uid."""
        self.idm.user_identifier = 'uid'
        assert self.idm.user == 'uid'

    def test_query_user(self):
        """Test query user function."""
        assert self.idm.query_user(self.idm.user) == 'query common_name'

    def test_logout_user(self):
        """Test logout user function."""
        assert self.idm.logout_user(self.idm.user) == 'logout common_name'

    def test_list(self):
        """Test list function."""
        assert self.idm.list() == 'list'

    def test_run(self):
        """Test run function."""
        assert self.idm.run() == 'query common_name'


@pytest.mark.parametrize(
    'args, expected',
    (
        (
            ['query', '--uid', 'uid', '--cn', 'common_name'],
            'query common_name',
        ),
        (
            ['logout', '--uid', 'uid', '--cn', 'common_name'],
            'logout common_name',
        ),
        (['list'], 'list'),
    ),
)
def test_logoutd_run_command_argument(args, expected):
    """Test the result of the run method with different command arguments."""
    idm = ConcreteLogoutdBase(args)
    assert idm.run() == expected


def test_run_logoutd_invalid_command():
    """Test invalid command."""
    with pytest.raises(SystemExit):
        ConcreteLogoutdBase(['invalid'])



"""File I/O module."""
import fcntl
import logging

from contextlib import contextmanager
from datetime import timedelta
from os import PathLike
from typing import Generator, IO

from wmflib.decorators import retry
from wmflib.exceptions import WmflibError


logger = logging.getLogger(__name__)


class FileIOError(WmflibError):
    """Custom exception class for errors of the BlockInFile class."""


class LockError(FileIOError):
    """Custom exception class raised when unable to exclusively lock a file."""


@contextmanager
def locked_open(file_path: PathLike, file_mode: str = 'r', *, timeout: int = 10) -> Generator[IO, None, None]:
    """Context manager to open a file with an exclusive lock on it and a retry logic.

    Examples:
        ::

            from wmflib.fileio import locked_open
            with locked_open('existing.file') as f:
                text = f.read()

            with locked_open('new.out', 'w') as f:
                f.write('Some text')

    Arguments:
        file_path (os.PathLike): the file path to open.
        file_mode (str, optional): the mode in which the file is opened, see :py:func:`open` for details.
        timeout (int, optional): the total timeout in seconds to wait to acquire the exclusive lock before giving up.
            Ten tries will be attempted to acquire the lock within the timeout.

    Raises:
        wmflib.fileio.LockError: on failure to acquire the exclusive lock on the file.

    Yields:
        file object: the open file with an exclusive lock on it.

    """
    tries = 10
    with open(file_path, file_mode, encoding='utf-8') as fd:
        try:
            # Decorate the call to the locking function to retry acquiring the lock:
            # decorator(decorator_args)(function)(function_args)
            # no-value-for-parameter is needed because pylint is confused by @ensure_wraps
            retry(  # pylint: disable=no-value-for-parameter
                tries=tries,
                delay=timedelta(seconds=timeout / tries),
                backoff_mode='constant',
                exceptions=(OSError, BlockingIOError)
            )(fcntl.flock)(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            logger.debug('Acquired exclusive lock on %s', file_path)
        except OSError as e:
            raise LockError(f'Unable to acquire exclusive lock on {file_path}') from e

        try:
            yield fd
        finally:
            fcntl.flock(fd, fcntl.LOCK_UN)
            logger.debug('Released exclusive lock on %s', file_path)



"""Actions module."""
import logging

from typing import Hashable, List


logger = logging.getLogger(__name__)


class Actions:
    """Class to keep track and log a set of actions performed and their result with a nice string representation."""

    def __init__(self, name: Hashable):
        """The instance gets initialized with the given name, that can represent a host or any other identifier.

        It allows to save actions performed on the given name with their status (success, warning, failure).
        It exposes the following public properties:

        - ``name``: the name passed to the instance at instantiation time.

        - ``has_warnings``: a :py:class:`bool` that is :py:data:`True` when at least one warning action was registered,
          :py:data:`False` otherwise.

        - ``has_failures``: a :py:class:`bool` that is :py:data:`True` when at least one failed action was registered,
          :py:data:`False` otherwise.

        When converted to a string it returns a nicely formatted representation of the instance and all its actions
        that can be used as-is in Phabricator::

            <name> (**<status>**)
              - <action1>
              - <action2>
              - ...

        The status is enclosed in double asterisks to be rendered as bold in Phabricator.

        Examples:
            ::

                actions = Actions('host1001')
                actions.success('Downtimed on Icinga')
                actions.success('Restarted ntp')
                print(actions)

            The above code will print::

                host1001 (**PASS**)
                  - Downtimed on Icinga
                  - Restarted ntp

        Arguments:
            name (typing.Hashable): the name of the set of actions to be registered.

        """
        self.name = name
        self.actions: List[str] = []
        self.has_warnings = False
        self.has_failures = False

    def __str__(self) -> str:
        """Custom string representation of the actions performed.

        Returns:
            str: the string representation.

        """
        actions = '\n'.join(f'  - {action}' for action in self.actions)
        return f'{self.name} (**{self.status}**)\n{actions}'

    @property
    def status(self) -> str:
        """Return the current status of the actions based on the worst result recorded.

        Returns:
            str:
                * ``FAIL``: if there was at least one failure action registered
                * ``WARN`` if there was at least one warning action and no ``FAIL`` actions registered
                * ``PASS`` if only success actions were registered

        """
        if self.has_failures:
            return 'FAIL'
        if self.has_warnings:
            return 'WARN'

        return 'PASS'

    def success(self, message: str) -> None:
        """Register a successful action, it gets also logged with info level.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.INFO, message)

    def failure(self, message: str) -> None:
        """Register a failed action, it gets also logged with error level.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.ERROR, message)
        self.has_failures = True

    def warning(self, message: str) -> None:
        """Register an action that require some attention, it gets also logged with warning level.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.WARNING, message)
        self.has_warnings = True

    def _action(self, level: int, message: str) -> None:
        """Register a generic action.

        Arguments:
            level (int): a logging level integer to register the action for.
            message (str): the action description.

        """
        logger.log(level, message)
        self.actions.append(message)


class ActionsDict(dict):
    """Custom dictionary with defaultdict capabilities for the :py:class:`wmflib.actions.Action` class.

    Automatically instantiate and returns a new instance of the :py:class:`wmflib.actions.Actions` class for every
    missing key like a :py:class:`collections.defaultdict`.

    When accessing a missing key, the key itself is passed to the new :py:class:`wmflib.actions.Actions` instance
    as ``name``.

    When converted to string returns a nicely formatted representation of the instance and all its items, that can be
    used as-is in Phabricator.

    Examples:
        ::

            actions = ActionsDict()
            actions['host1001'].success('Downtimed on Icinga')
            actions['host1001'].failure('**Failed to restart ntp**')  # Will be rendered in bold in Phabricator
            actions['host2001'].warning('//Host with alerts on Icinga//')  # Will be rendered in italic in Phabricator
            print(actions)

        The above code will print::

            - host1001 (**FAIL**)
              - Downtimed on Icinga
              - **Failed to restart ntp**

            - host2001 (**WARN**)
              - //Host not optimal on Icinga//

        It will also include a final newline at the end of the block that doesn't get rendered in this documentation.

    """

    def __missing__(self, key: Hashable) -> Actions:
        """Instantiate a new Actions instance for the missing key like a defaultdict.

        Parameters as required by Python's data model, see :py:method:`object.__missing__`.

        """
        self[key] = Actions(key)
        return self[key]

    def __str__(self) -> str:
        """Custom string representation of the instance.

        Returns:
            str: the string representation of each item in the dictionary, newline-separated.

        """
        return '\n'.join(f'- {value}\n' for value in self.values())



"""Pytest customization for unit tests."""
import fcntl
import uuid

import pytest


@pytest.fixture
def locked_file(tmp_path):
    """Pytest fixture that creates a random file and locks it exclusively."""
    tmp_file = tmp_path / str(uuid.uuid4())
    with open(tmp_file, 'w', encoding='utf-8') as fd:
        fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        yield tmp_file
        fcntl.flock(fd, fcntl.LOCK_UN)



"""Constants module."""


ALL_DATACENTERS = ('eqiad', 'codfw', 'esams', 'ulsfo', 'eqsin', 'drmrs')
"""tuple: all WMF datacenters."""


CORE_DATACENTERS = ("eqiad", "codfw")
"""tuple: WMF core datacenters."""


PUBLIC_AUTHDNS = ('208.80.154.238', '208.80.153.231', '91.198.174.239')
"""tuple: publicly Authorative DNS servers for the WMF domains."""


DATACENTER_NUMBERING_PREFIX = {
    'eqiad': '1',
    'codfw': '2',
    'esams': '3',
    'ulsfo': '4',
    'eqsin': '5',
    'drmrs': '6',
}
"""dict: mapping of hostname prefix for numbered servers for each WMF datacenter."""



"""Decorators module."""
import logging
import time

from dataclasses import dataclass
from datetime import timedelta
from functools import wraps
from typing import Any, Callable, cast, Dict, Optional, Tuple, Type, Union

from wmflib.exceptions import WmflibError


logger = logging.getLogger(__name__)


# TODO: use TypedDict once Python 3.7 support is removed
@dataclass
class RetryParams:
    """Retry decorator parameters class.

    This class represents the parameters that the :py:func:`wmflib.decorators.retry` decorator accepts and allow
    to validate them.

    It is also used for the `dynamic_params_callbacks` parameter of the same retry decorator to allow the callback to
    safely modify the decorator's parameters at runtime.
    """

    tries: int
    delay: timedelta
    backoff_mode: str
    exceptions: Tuple[Type[Exception], ...]
    failure_message: str

    def validate(self) -> None:
        """Validate the consistency of the current values of the instance properties.

        Raises:
            wmflib.exceptions.WmflibError: if any field has an invalid value.

        """
        if self.backoff_mode not in ('constant', 'linear', 'power', 'exponential'):
            raise WmflibError(f'Invalid backoff_mode: {self.backoff_mode}')

        if self.backoff_mode == 'exponential' and self.delay.total_seconds() < 1:
            raise WmflibError(f'Delay must be greater than 1 if backoff_mode is exponential, got {self.delay}')

        if self.tries < 1:
            raise WmflibError(f'Tries must be a positive integer, got {self.tries}')

        if not self.failure_message:
            raise WmflibError('A failure_message must be set.')


def ensure_wrap(func: Callable) -> Callable:
    """Decorator to wrap other decorators to allow to call them both with and without arguments.

    Arguments:
        func: the decorated function, it must be a decorator. A decorator that accepts only one positional argument
            that is also a callable is not supported.

    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Callable:
        """Decorator wrapper."""
        if len(args) == 1 and not kwargs and callable(args[0]):  # Called without arguments
            return func(args[0])

        return lambda real_func: func(real_func, *args, **kwargs)  # Called with arguments

    return wrapper


@ensure_wrap
def retry(
    func: Callable,
    *,
    tries: int = 3,
    delay: timedelta = timedelta(seconds=3),
    backoff_mode: str = 'exponential',
    exceptions: Tuple[Type[Exception], ...] = (WmflibError,),
    failure_message: Optional[str] = None,
    dynamic_params_callbacks: Tuple[Callable[[RetryParams, Callable, Tuple, Dict[str, Any]], None], ...] = (),
) -> Callable:
    """Decorator to retry a function or method if it raises certain exceptions with customizable backoff.

    Note:
        The decorated function or method must be idempotent to avoid unwanted side effects.
        It can be called with or without arguments, in the latter case all the default values will be used.

    Examples:
        Define a function that polls for the existence of a file, retrying with the default parameters::

            >>> from pathlib import Path
            >>> from wmflib.decorators import retry
            >>> from wmflib.exceptions import WmflibError
            >>> @retry
            ... def poll_file(path: Path):
            ...     if not path.exists():
            ...         raise WmflibError(f'File {path} not found')
            ...
            >>> poll_file(Path('/tmp'))
            >>> poll_file(Path('/tmp/nonexistent'))
            [1/3, retrying in 3.00s] Attempt to run '__main__.poll_file' raised: File /tmp/nonexistent not found
            [2/3, retrying in 9.00s] Attempt to run '__main__.poll_file' raised: File /tmp/nonexistent not found
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
              File "/usr/lib/python3/dist-packages/wmflib/decorators.py", line 160, in wrapper
                return func(*args, **kwargs)
              File "<stdin>", line 4, in poll_file
            wmflib.exceptions.WmflibError: File /tmp/nonexistent not found
            >>>

        Same example, but customizing the decorator parameters::

            >>> from datetime import timedelta
            >>> from pathlib import Path
            >>> from wmflib.decorators import retry
            >>> @retry(tries=5, delay=timedelta(seconds=30), backoff_mode='constant', failure_message='File not found',
            ...        exceptions=(RuntimeError,))
            ... def poll_file(path: Path):
            ...     if not path.exists():
            ...         raise RuntimeError(path)
            ...
            [1/5, retrying in 30.00s] File not found: /tmp/nonexistent
            [2/5, retrying in 30.00s] File not found: /tmp/nonexistent
            [3/5, retrying in 30.00s] File not found: /tmp/nonexistent
            [4/5, retrying in 30.00s] File not found: /tmp/nonexistent
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
              File "/usr/lib/python3/dist-packages/wmflib/decorators.py", line 160, in wrapper
                return func(*args, **kwargs)
              File "<stdin>", line 5, in poll_file
            RuntimeError: /tmp/nonexistent

    Arguments:
        func (function, method): the decorated function.
        tries (int, optional): the number of times to try calling the decorated function or method before giving up.
            Must be a positive integer.
        delay (datetime.timedelta, optional): the initial delay for the first retry, used also as the base for the
            backoff algorithm.
        backoff_mode (str, optional): the backoff mode to use for the delay, available values are::

            constant:    delay       => 3, 3,  3,  3,   3, ...;
            linear:      delay * N   => 3, 6,  9, 12,  15, ...; N in [1, tries]
            power:       delay * 2^N => 3, 6, 12, 24,  48, ...; N in [0, tries - 1]
            exponential: delay^N     => 3, 9, 27, 81, 243, ...; N in [1, tries], delay must be > 1.

        exceptions (type, tuple, optional): the decorated function call will be retried if it fails until it succeeds
            or `tries` attempts are reached. A retryable failure is defined as raising any of the exceptions listed.
        failure_message (str, optional): the message to log each time there's a retryable failure. Retry information
            and exception message are also included. Default: "Attempt to run '<fully qualified function>' raised"
        dynamic_params_callbacks (tuple): a tuple of callbacks that will be called at runtime to allow to modify the
            decorator's parameters. Each callable must adhere to the following interface::

                def adjust_some_parameter(retry_params: RetryParams, func: Callable, args: Tuple, kwargs: Dict) -> None
                    # Modify the retry_params parameter possibly using the decorated function object or its parameters
                    # that are passed as tuple for the positional arguments and a dictionary for the keyword arguments

            This is a practical example that defines a callback that doubles the delay parameter of the ``@retry``
            decorator if the decorated function/method has a 'slow' keyword argument that is to True::

                def double_delay(retry_params, func, args, kwargs):
                    if kwargs.get('slow', False):
                        retry_params.delay = retry_params.delay * 2

                @retry(delay=timedelta(seconds=10), dynamic_params_callbacks=(double_delay,))
                def do_something(slow=False):
                    # This method will be retried using 10 seconds as delay parameter in the @retry decorator, but
                    # if the 'slow' parameter is set to True it will use a delay of 20 seconds instead.
                    # Do something here.

            **While the callbacks will have access to the parameters passed to the decorated function, they should be
            treated as read-only variables.**

    Returns:
        function: the decorated function.

    """
    if not failure_message:
        failure_message = f"Attempt to run '{func.__module__}.{func.__qualname__}' raised"

    static_params: Dict[str, Any] = {
        'tries': tries,
        'delay': delay,
        'backoff_mode': backoff_mode,
        'exceptions': exceptions,
        'failure_message': failure_message,
    }

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Decorated function."""
        params = RetryParams(**static_params)
        for dynamic_params_callback in dynamic_params_callbacks:
            dynamic_params_callback(params, func, args, kwargs)

        params.validate()
        attempt = 0
        while attempt < params.tries - 1:
            attempt += 1
            try:
                # Call the decorated function or method
                return func(*args, **kwargs)
            except exceptions as e:
                sleep = get_backoff_sleep(params.backoff_mode, params.delay.total_seconds(), attempt)
                logger.warning("[%d/%d, retrying in %.2fs] %s: %s",
                               attempt, params.tries, sleep, params.failure_message, _exception_message(e))
                time.sleep(sleep)

        return func(*args, **kwargs)

    return wrapper


def _exception_message(exception: BaseException) -> str:
    """Joins the message of the given exception with those of any chained exceptions.

    Arguments:
        exception (BaseException): The most-recently raised exception.

    Returns:
        str: The joined message, formatted suitably for logging.

    """
    message_parts = [str(exception)]
    while exception.__cause__ is not None or exception.__context__ is not None:
        # __cause__ and __context__ shouldn't both be set, but we use the same logic here as the built-in
        # exception handler, giving __cause__ priority, as described in PEP 3134. We list messages in
        # reverse order from the built-in handler (i.e. newest exception first) since we aren't following a
        # traceback.
        if exception.__cause__ is not None:
            message_parts.append(f'Caused by: {exception.__cause__}')
            exception = exception.__cause__
        else:  # e.__context__ is not None, due to the while condition.
            message_parts.append(f'Raised while handling: {exception.__context__}')
            exception = cast(BaseException, exception.__context__)  # Casting away the Optional.
    return '\n'.join(message_parts)


def get_backoff_sleep(backoff_mode: str, base: Union[int, float], index: int) -> Union[int, float]:
    """Calculate the amount of sleep for this attempt.

    Arguments:
        backoff_mode (str): the backoff mode to use for the delay, see the documentation for retry().
        base (int, float): the base for the backoff algorithm.
        index (int): the index to calculate the Nth sleep time for the backoff.

    Return:
        int, float: the amount of sleep to perform for the backoff.

    """
    if backoff_mode == 'constant':
        sleep = base
    elif backoff_mode == 'linear':
        sleep = base * index
    elif backoff_mode == 'power':
        sleep = base * 2 ** (index - 1)
    elif backoff_mode == 'exponential':
        sleep = base ** index
    else:
        raise ValueError(f'Invalid backoff_mode: {backoff_mode}')

    return sleep



"""Interactive module."""
import getpass
import logging
import os
import sys

from typing import Any, Callable, Sequence

from wmflib.exceptions import WmflibError


logger = logging.getLogger(__name__)
MIN_SECRET_SIZE: int = 6


class InputError(WmflibError):
    """Custom exception class raised on invalid input from the user in interactive mode."""


class AbortError(WmflibError):
    """Custom exception class raised when an action is manually aborted."""


def ask_input(message: str, choices: Sequence[str]) -> str:
    """Ask the user for input in interactive mode.

    Examples:
        ::

            >>> choices = ['A', 'B']
            >>> response = ask_input(f'Choose a door between {choices}', choices)
            ==> Choose a door between ['A', 'B']
            > a
            ==> Invalid response, please type one of: A,B. After 3 wrong answers the task will be aborted.
            > A
            >>> response
            'A'

    Arguments:
        message (str): the message to be printed before asking for confirmation.
        choices (sequence): the available choices of possible answers that the user can give. Values must be strings.

    Returns:
        str: the selected choice.

    Raises:
        wmflib.interactive.InputError: if not in a TTY or on too many invalid answers.

    """
    prefix = '\x1b[36m==>\x1b[39m'  # Cyan ==> prefix
    if not sys.stdout.isatty():
        raise InputError('Not in a TTY, unable to ask for input')

    print(f'{prefix} {message}')

    for _ in range(3):
        response = input('> ')
        if response in choices:
            return response

        print(f'{prefix} Invalid response, please type one of: {",".join(choices)}. '
              'After 3 wrong answers the task will be aborted.')

    raise InputError('Too many invalid answers')


def ask_confirmation(message: str) -> None:
    """Ask the use for confirmation in interactive mode.

    Examples:
        ::

            >>> ask_confirmation('Ready to continue?')
            ==> Ready to continue?
            Type "go" to proceed or "abort" to interrupt the execution
            > go
            >>> ask_confirmation('Ready to continue?')
            ==> Ready to continue?
            Type "go" to proceed or "abort" to interrupt the execution
            > abort
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
              File "/usr/lib/python3/dist-packages/wmflib/interactive.py", line 69, in ask_confirmation
                raise AbortError('Confirmation manually aborted')
            wmflib.interactive.AbortError: Confirmation manually aborted

    Arguments:
        message (str): the message to be printed before asking for confirmation.

    Raises:
        wmflib.interactive.InputError: if not in a TTY or on too many invalid answers.
        wmflib.interactive.AbortError: if manually aborted.

    """
    response = ask_input('\n'.join((message, 'Type "go" to proceed or "abort" to interrupt the execution')),
                         ['go', 'abort'])
    if response == 'abort':
        raise AbortError('Confirmation manually aborted')


def confirm_on_failure(func: Callable, *args: Any, **kwargs: Any) -> Any:
    """Execute a function asking for confirmation to retry, abort or skip.

    Examples:
        ::

            >>> def test(fail=False):
            ...     if fail:
            ...         raise RuntimeError('Failed')
            ...
            >>> confirm_on_failure(test)
            >>> confirm_on_failure(test, fail=True)
            Failed to run __main__.test: Failed
            ==> What do you want to do? "retry" the last command, manually fix the issue and "skip" the last command to
                continue the execution or completely "abort" the execution.
            > retry
            Failed to run __main__.test: Failed
            ==> What do you want to do? "retry" the last command, manually fix the issue and "skip" the last command to
                continue the execution or completely "abort" the execution.
            > skip
            >>>

    Arguments:
        func (callable): the function/method to execute.
        *args (mixed): all the positional arguments to pass to the function/method.
        *kwargs (mixed): all the keyword arguments to pass to the function/method.

    Returns:
        mixed: what the called function returns, or :py:data:`None` if the execution should continue skipping this
        step because has been manually fixed.

    Raises:
        wmflib.interactive.AbortError: on manually aborted tasks.

    """
    retry_message = ('What do you want to do? "retry" the last command, manually fix the issue and "skip" the last '
                     'command to continue the execution or completely "abort" the execution.')
    choices = ['retry', 'skip', 'abort']

    while True:
        try:
            ret = func(*args, **kwargs)
        except AbortError:
            raise
        except BaseException as e:  # pylint: disable=broad-except
            logger.error('Failed to run %s.%s: %s', func.__module__, func.__qualname__, e)
            logger.debug('Traceback', exc_info=True)
            response = ask_input(retry_message, choices)
            if response == 'skip':
                return None
            if response == 'abort':
                raise AbortError('Task manually aborted') from e
        else:
            return ret


def get_username() -> str:
    """Detect and return the name of the effective running user even if run as root.

    Examples:
        ::

            >>> get_username()
            'user'

    Returns:
        str: the name of the effective running user or ``-`` if unable to detect it.

    """
    user = os.getenv('USER')
    sudo_user = os.getenv('SUDO_USER')

    if sudo_user is not None and sudo_user != 'root':
        return sudo_user

    if user is not None:
        return user

    return '-'


def ensure_shell_is_durable() -> None:
    """Ensure it is running either in non-interactive mode or in a screen/tmux session, raise otherwise.

    Examples:
        ::

            >>> ensure_shell_is_durable()  # Will raise if not in a tmux/screen session
            >>>

    Raises:
        wmflib.exceptions.WmflibError: if in a non-durable shell session.

    """
    # STY is for screen, TMUX is for tmux. Not using `getenv('NAME') is not None` to check they are not empty.
    # TODO: verify if the check on TERM is redundant.
    if (sys.stdout.isatty() and not os.getenv('STY', '') and not os.getenv('TMUX', '')
            and 'screen' not in os.getenv('TERM', '') and 'tmux' not in os.getenv('TERM', '')):
        raise WmflibError('Must be run in non-interactive mode or inside a screen or tmux.')


def get_secret(title: str, *, confirm: bool = False) -> str:
    """Ask the user for a secret e.g. password.

    Examples:
        ::

            >>> secret = get_secret('Secret key')
            Secret key:
            Secret must be at least 6 characters. try again:
            >>> secret = get_secret('Secret key', confirm=True)  # Will raise if the confirmation doesn't match
            Secret key:
            Again, just to be sure:
            >>>

    Arguments:
        title (str): The message to show the user.
        confirm (bool, optional): If :py:data:`True` ask the user to confirm the password.

    Returns:
        str: the secret.

    Raises:
        wmflib.exceptions.WmflibError: if the password confirmation does not match and confirm is :py:data:`True`.

    """
    new_secret = getpass.getpass(prompt=f'{title}: ')

    while len(new_secret) < MIN_SECRET_SIZE:
        new_secret = getpass.getpass(
            prompt=f'Secret must be at least {MIN_SECRET_SIZE} characters. try again: ')

    if confirm and new_secret != getpass.getpass(prompt='Again, just to be sure: '):
        raise WmflibError(f'{title}: Passwords did not match')

    return new_secret



"""Requests module tests."""
from unittest import mock

import pytest

from requests import Request, Session

from wmflib import requests


def test_timeout_http_adapter_init_default():
    """If instantiated without any parameter the TimeoutHTTPAdapter should set the default timeout."""
    adapter = requests.TimeoutHTTPAdapter()
    assert adapter.timeout == requests.DEFAULT_TIMEOUT


def test_timeout_http_adapter_init_custom():
    """The TimeoutHTTPAdapter should use the given timeout, if specified."""
    adapter = requests.TimeoutHTTPAdapter(timeout=9.9)
    assert adapter.timeout == 9.9


@pytest.mark.parametrize('timeout, expected', (
    (None, requests.DEFAULT_TIMEOUT),
    (1.0, 1.0),
))
@mock.patch('wmflib.requests.HTTPAdapter.send', spec_set=True)
def test_timeout_http_adapter_send_default(mocked_send, timeout, expected):
    """Sending a request with the TimeoutHTTPAdapter should use the default timeout if not set."""
    adapter = requests.TimeoutHTTPAdapter()
    request = Request('GET', 'https://example.com/')
    prepared_request = request.prepare()
    if timeout:
        adapter.send(prepared_request, timeout=timeout)
    else:
        adapter.send(prepared_request)

    assert mocked_send.call_count == 1
    assert mocked_send.mock_calls[0][2].get('timeout') == expected


def test_session():
    """Calling session should return a Requests's session pre-configured."""
    session = requests.http_session('UA-name')
    assert isinstance(session, Session)
    assert 'UA-name' in session.headers['User-Agent']



"""DNS module."""
import logging

from typing import cast, List, Optional, Sequence, Union

from dns import resolver, reversename, rrset
from dns.exception import DNSException
from dns.name import Name

from wmflib.constants import PUBLIC_AUTHDNS
from wmflib.exceptions import WmflibError


logger = logging.getLogger(__name__)


class DnsError(WmflibError):
    """Custom exception class for errors of the Dns class."""


class DnsNotFound(DnsError):
    """Custom exception class to indicate the record was not found.

    One or more resource records might exist for this domain but no record matches the resource record type requested.
    """


class Dns:
    """Class to interact with the DNS."""

    def __init__(self, *, nameserver_addresses: Optional[Sequence] = None, port: Optional[int] = None) -> None:
        """Initialize the instance optionally specifying the nameservers to use.

        Examples:
            Using the host's default DNS resolvers::

                >>> from wmflib.dns import Dns
                >>> dns = Dns()

            Using a specific set of resolvers and port::

                >>> from wmflib.dns import Dns
                >>> dns = Dns(nameserver_addresses=['10.0.0.1', '10.0.0.2'], port=5353)

        Arguments:
            nameserver_addresses (Sequence, optional): the nameserveres address to use, if not set uses the OS
                configuration.
            port (int, optional): the port the ``nameserver_addresses`` nameserveres is listening to, if different from
                the default 53. This applies only if a nameserveres is explicitelyes specified.

        """
        if nameserver_addresses is not None:
            self._resolver = resolver.Resolver(configure=False)
            if port is not None:
                self._resolver.port = port
            self._resolver.nameservers = nameserver_addresses
        else:
            self._resolver = resolver.Resolver()

    def resolve_ipv4(self, name: str) -> List[str]:
        """Perform a DNS lookup for an A record for the given name.

        Examples:
            ::

                >>> dns.resolve_ipv4('api.svc.eqiad.wmnet')
                ['10.2.2.22']

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv4 addresses as strings returned by the DNS response.

        """
        return self._resolve_addresses(name, 'A')

    def resolve_ipv6(self, name: str) -> List[str]:
        """Perform a DNS lookup for an AAAA record for the given name.

        Examples:
            ::

                >>> dns.resolve_ipv6('wikimedia.org')
                ['2620:0:861:ed1a::1']

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv6 addresses as strings returned by the DNS response.

        """
        return self._resolve_addresses(name, 'AAAA')

    def resolve_ips(self, name: str) -> List[str]:
        """Perform a DNS lookup for A and AAAA records for the given name.

        Examples:
            ::

                >>> dns.resolve_ips('wikimedia.org')
                ['208.80.154.224', '2620:0:861:ed1a::1']

        Arguments:
            name (str): the name to resolve.

        Returns:
            list: the list of IPv4 and IPv6 addresses as strings returned by the DNS response.

        Raises:
            wmflib.dns.DnsNotFound: when no address is found.

        """
        addresses = []
        for func in ('resolve_ipv4', 'resolve_ipv6'):
            try:
                addresses += getattr(self, func)(name)
            except DnsNotFound:
                pass  # Allow single stack answers

        if not addresses:
            raise DnsNotFound(f'Record A or AAAA not found for {name}')

        return addresses

    def resolve_ptr(self, address: str) -> List[str]:
        """Perform a DNS lookup for PTR record for the given address.

        Examples:
            ::

                >>> dns.resolve_ptr('208.80.154.224')
                ['text-lb.eqiad.wikimedia.org']

        Arguments:
            address (str): the IPv4 or IPv6 address to resolve.

        Returns:
            list: the list of absolute target PTR records as strings, without the trailing dot.

        """
        response = self.resolve(reversename.from_address(address), 'PTR')
        return self._parse_targets(cast(rrset.RRset, response.rrset))

    def resolve_cname(self, name: str) -> str:
        """Perform a DNS lookup for CNAME record for the given name.

        Examples:
            ::

                >>> dns.resolve_cname('puppet.codfw.wmnet')
                'puppetmaster2001.codfw.wmnet'

        Arguments:
            name (str): the name to resolve.

        Returns:
            str: the absolute target name for this CNAME, without the trailing dot.

        """
        targets = self._parse_targets(cast(rrset.RRset, self.resolve(name, 'CNAME').rrset))
        if len(targets) != 1:
            raise DnsError(f'Found multiple CNAMEs target for {name}: {targets}')

        return targets[0]

    def resolve(self, qname: Union[str, Name], record_type: str) -> resolver.Answer:
        """Perform a DNS lookup for the given qname and record type.

        Examples:
            ::

                >>> response = dns.resolve('wikimedia.org', 'MX')
                >>> [rdata.to_text() for rdata in response.rrset]
                ['10 mx1001.wikimedia.org.', '50 mx2001.wikimedia.org.']

        Arguments:
            qname (str): the name or address to resolve.
            record_type (str): the DNS record type to lookup for, like 'A', 'AAAA', 'PTR', etc.

        Returns:
            dns.resolver.Answer: the DNS response.

        Raises:
            wmflib.dns.DnsNotFound: if there are no records for the given record type but the qname has records for
                different record type(s).
            wmflib.dns.DnsError: on generic error.

        """
        try:
            response = self._resolver.query(qname, record_type)
            logger.debug('Resolved %s record for %s: %s', record_type, qname, response.rrset)
        except (resolver.NoAnswer, resolver.NXDOMAIN) as e:
            raise DnsNotFound(f'Record {record_type} not found for {qname}') from e
        except DNSException as e:
            raise DnsError(f'Unable to resolve {record_type} record for {qname}') from e

        return response

    def _resolve_addresses(self, name: str, record_type: str) -> List[str]:
        """Extract and return all the matching addresses for the given name and record type.

        Arguments:
            name (str): the name to resolve.
            record_type (str): the DNS record type to lookup for, like 'A' and 'AAAA'.

        Returns:
            list: the list of IPv4 or IPv6 addresses as strings returned by the DNS response.

        """
        return [rdata.address for rdata in cast(rrset.RRset, self.resolve(name, record_type).rrset)]

    @staticmethod
    def _parse_targets(response_set: rrset.RRset) -> List[str]:
        """Extract and return all the matching names from the given rrset without the trailing dot.

        Arguments:
            response_set (dns.rrset.RRset): the RRset to parse.

        Returns:
            list: the list of absolute target record names as strings without the trailing dot.

        Raises:
            wmflib.dns.DnsError: if a relative record is found.

        """
        targets = []
        for rdata in response_set:
            target = rdata.target.to_text()
            if target[-1] != '.':
                raise DnsError(f'Unsupported relative target {target} found')

            targets.append(target[:-1])

        return targets


class PublicAuthDns(Dns):
    """Class to interact with the DNS using the wikimedia foundation authoritative servers."""

    def __init__(self) -> None:
        """Initialize the instance with the WMF public authoritative namerservers.

        It uses the nameservers defined in :py:const:`wmflib.constants.PUBLIC_AUTHDNS`.

        Examples:
            ::

                >>> from wmflib.dns import PublicAuthDns
                >>> dns = PublicAuthDns()

        """
        super().__init__(nameserver_addresses=PUBLIC_AUTHDNS)



"""File I/O module tests."""
import fcntl

from unittest import mock

import pytest

from wmflib.fileio import locked_open, LockError


def try_lock_file(test_file):
    """Try to get an exclusive lock in the given file. Return True on success, False on failure."""
    with open(test_file, encoding='utf-8') as fd:
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            fcntl.flock(fd, fcntl.LOCK_UN)
            return True
        except OSError:
            return False


@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_locked_open_success(mocked_sleep, tmp_path):
    """It should acquire an exclusive lock to a file and write to it."""
    test_file = tmp_path / 'acquire_lock'
    with locked_open(test_file, 'w') as fd:
        fd.write('some text')
        assert not try_lock_file(test_file)

    assert try_lock_file(test_file)
    assert test_file.read_text() == 'some text'
    mocked_sleep.assert_not_called()


@mock.patch('wmflib.decorators.time.sleep', return_value=None)
def test_locked_open_fail(mocked_sleep, locked_file):
    """It should retry to get an exclusive lock and raise a LockError exception on failure."""
    with pytest.raises(LockError, match='Unable to acquire exclusive lock on'):
        with locked_open(locked_file):
            assert False, 'Execution should not reach this point'

    mocked_sleep.assert_has_calls([mock.call(1.0)] * 9)  # 10 tries, 9 sleeps



"""IRC module."""

import logging
import socket

from typing import Tuple


class SocketHandler(logging.Handler):
    """Log handler to communicate with logmsgbot.

    Sends log events to a tcpircbot server for relay to an IRC channel.
    The combination of hostname:port identifies a specific IRC channel,
    and for the moment only the #wikimedia-operations one is configured.
    Please also keep in mind that SRE allows only specific hostnames to
    send traffic to tcpircbot, so follow up with them in case you have
    doubts or special requests.
    For more info, please check the tcpircbot config in puppet.

    Examples:
        ::

            >>> import logging
            >>> from wmflib.interactive import get_username
            >>> from wmflib.irc import SocketHandler
            >>> irc_logger = logging.getLogger('irc_logger')
            >>> irc_logger.setLevel(logging.INFO)
            >>> irc_logger.addHandler(SocketHandler('irchost.domain', 1234, get_username()))
            >>> irc_logger.info('Message')

        The tcpircbot that received the message will then send it to the configured IRC channel::

            logmsgbot: user@host1001.domain Message

    """

    prefix = '{s.username}@{s.hostname}'

    def __init__(self, host: str, port: int, username: str) -> None:
        """Initialize the IRC socket handler.

        Arguments:
            host (str): tcpircbot hostname.
            port (int): tcpircbot listening port.
            username (str): the user to refer in the IRC messages.

        """
        super().__init__()
        self.addr: Tuple[str, int] = (host, port)
        self.username = username
        self.level = logging.INFO
        self.hostname = socket.gethostname()

    def _send_message(self, message: str, record: logging.LogRecord) -> None:
        """Send a custom message string on a socket."""
        sock = None
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1.0)
            sock.connect(self.addr)
            sock.sendall(message.encode('utf-8'))
        except OSError:
            self.handleError(record)
        finally:
            if sock is not None:
                sock.close()

    def emit(self, record: logging.LogRecord) -> None:
        """According to Python logging.Handler interface.

        See https://docs.python.org/3/library/logging.html#handler-objects
        """
        message = self.prefix.format(s=self) + ' ' + record.getMessage()
        self._send_message(message, record)


class SALSocketHandler(SocketHandler):
    """Log handler to !log on a SAL.

    See the parent class for usage examples.
    """

    # Stashbot expects !log messages relayed by logmsgbot to have the
    # format: "!log <nick> <msg>". The <nick> is parsed out and used as
    # the label of who actually made the SAL entry.
    prefix = '!log {s.username}@{s.hostname}'



"""Tests package for wmflib."""

import logging

from pathlib import Path

import pytest

from pkg_resources import parse_version

CAPLOG_MIN_VERSION = '3.3.0'
TESTS_BASE_PATH = Path(__file__).parent.resolve()


def get_fixture_path(*paths):
    """Return the absolute path of the given fixture.

    Arguments:
        *paths: arbitrary positional arguments used to compose the absolute path to the fixture.

    Returns:
        str: the absolute path of the selected fixture.

    """
    return Path(TESTS_BASE_PATH, 'fixtures', *paths)


require_caplog = pytest.mark.skipif(
    parse_version(pytest.__version__) < parse_version(CAPLOG_MIN_VERSION), reason='Requires caplog fixture')


def check_logs(logs, message, level):
    """Assert that a log record with the given message and level is present."""
    for record in logs.records:
        if message in record.getMessage():
            assert record.levelno == level
            break
    else:
        raise RuntimeError(f"{logging.getLevelName(level)} log record with message '{message}' not found")



"""Base module IDM related classes and functions."""

from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from logging import getLogger
from typing import List, Optional

from wmflib.exceptions import WmflibError


class IdmValueError(WmflibError):
    """Raised by the IDM module value errors."""


# We make the description optional to deal with the following issue
# https://github.com/python/mypy/issues/9170
def logoutd_args(description: Optional[str] = None, args: Optional[List] = None) -> Namespace:
    """Logout scripts common CLI for parsing the command line arguments.

    When not using the higher level API :py:class:`wmflib.idm.LogoutdBase`, a user could just implement their own
    script still using the same command line arguments processing, of the form::

        from wmflib.idm import logoutd_args
        args = logoutd_args('Some description')
        # write your own script

    Arguments:
        description (str): the description to use
        args (list): A list of arguments to use (used for testing)

    Returns:
        `argparse.Namespace`: The parsed argparser Namespace

    """
    if description is None:
        raise IdmValueError('Must provide a string description')

    parser = ArgumentParser(description=description)
    parser.add_argument('-v', '--verbose', action='count', default=0)
    sub = parser.add_subparsers(dest='command', required=True)
    query = sub.add_parser('query', help='display the status of logged-in users')
    query.add_argument(
        '-u',
        '--uid',
        required=True,
        help='The uid of the user to use',
    )
    query.add_argument('-c', '--cn', required=True, help='The cn of the user to use')

    sub.add_parser(
        'logout',
        parents=[query],
        add_help=False,
        help='display the status of logged in users',
    )

    sub.add_parser('list', help='list all active sessions')

    return parser.parse_args(args)


class LogoutdBase(ABC):  # noqa: D300,D301 See https://github.com/PyCQA/pydocstyle/issues/542
    """Base class for a standardized API for logout scripts.

    A simple logout.d python script would then be of the form::

        from wmflib.idm import LogoutdBase

        class MyLogoutd(LogoutdBase):
            \"\"\"Some description.\"\"\"

            def logout_user(self, user):
                # logout the given user

            def query_user(self, user):
                # check the login status of a given user

            def list(self):
                # list all active login sessions

        logoutd = MyLogoutd()
        raise SystemExit(logoutd.run())  # This includes the parsing of command line arguments.

    """

    user_identifier: str = 'cn'

    def __init__(self, args: Optional[List] = None) -> None:
        """Init function.

        Arguments:
            args (list): A list of arguments to use (used for testing)

        """
        self._args = logoutd_args(self.__doc__, args)
        self._logger = getLogger('.'.join((self.__module__, self.__class__.__name__)))

    @property
    def user(self) -> str:
        """Return either common_name or uid with a preference for common_name.

        Returns:
            (str): representing the user

        """
        return getattr(self._args, self.user_identifier)

    @abstractmethod
    def logout_user(self, user: str) -> int:
        """Log out the specified user.

        Arguments:
            user (User): object representing the user

        Returns:
            (int): 0 if the users session was successfully cleared otherwise 1

        """

    @abstractmethod
    def query_user(self, user: str) -> int:
        """Return status of logged in user.

        Arguments:
            user (User): object representing the user

        Returns:
            (int): 1 if the user is logged in otherwise 0

        """

    @abstractmethod
    def list(self) -> int:
        """Return data of all logged in users.

        Returns:
            (int): 0 on success non-zero on fail

        """

    def run(self) -> int:
        """Execute the correct action.

        Returns:
            (int): exit code depends on command

        """
        self._logger.debug('Running action: %s', self._args.command)
        if self._args.command == 'query':
            return self.query_user(self.user)
        if self._args.command == 'logout':
            return self.logout_user(self.user)
        return self.list()
